/********************************************************************/
/*                              Poly                                */
/*                                                                  */
/*   Luc Maranget, projet Moscova, Inria Rocquencourt               */
/*                                                                  */
/* Copyright 2005 Institut National de Recherche en Informatique et */
/* en Automatique.  All rights reserved.  This file is distributed  */
/* under the terms of the Q Public License version 1.0.             */ 
/********************************************************************/

#include <stdio.h>
#include <ctype.h>
#include "count.h"

#define HIGHMASK (1UL << (8*sizeof(unsigned long)-1))



counter promote(unsigned long i)
{
  counter r;

  r.high = 0;
  r.low = i;
  return r;
}

counter add(counter x,counter y)
{
  unsigned long w;
  counter r;

  w = r.low = x.low + y.low;
  r.high = x.high + y.high;

  if (w < x.low || w < y.low)
    r.high++;

  return r;
}

counter mult2(counter x)
{
  unsigned long w;

  w = x.low << 1;
  x.high <<= 1;
  if (x.low != (w >> 1))
    x.high++;
  x.low = w;

  return x;
}

counter multi(counter x,unsigned long i)
{
  if (i == 0) {
    counter r = {0,0};

    return r;
  }
  if (i % 2 == 1)
    return add(x,mult2(multi(x,i/2)));
  else
    return mult2(multi(x,i/2));
}

int zero(counter x)
{
  return x.high == 0 && x.low == 0;
}


counter div2(counter x)
{
  x.low >>= 1;
  if (x.high % 2 == 1)
    x.low |= HIGHMASK;
  x.high >>= 1;
  return x;
}

/*
counter mult(counter x,counter y)
{
  if (zero(y)) {
    counter r = {0,0};

    return r;
  }
  if (y.low % 2 == 0) {
    y = div2(y);
    return add(x,mult2(mult(x,y)));
  } else {
    y = div2(y);
    return mult2(mult(x,y));
  }
}
*/
typedef struct {
  counter quo;
  unsigned rem;
} div;

div div10(counter x)
{
  unsigned long qh,rh,ql,rl;
  unsigned long w;
  counter w1 = {0,0};
  counter w2 = {0,0};
  div r;

  unsigned long rb,qb;

  /*  fprintf(stderr,"div10 {%lu, %lu} = ",x.high,x.low); */

  qb = HIGHMASK / 5;
  rb = HIGHMASK % 5;
  rb *= 2;

  qh = x.high/10 ; rh = x.high % 10;
  ql = x.low/10  ; rl = x.low % 10;

  w = rb * rh + rl;
  r.rem = w % 10;
  w /= 10;

  w1 = promote(w);
  ADD(w1,ql);

  w2 = promote(qh);
  w2 = multi(w2,rb);
  w1 = add(w1,w2);

  w2 = promote(qb);
  w1 = add(w1,multi(w2,rh));

  w2 = multi(w2,10);
  w2 = multi(w2,qh);
  r.quo = add(w1,w2);

  /*  fprintf(stderr,"{%lu, %lu} %u\n",r.quo.high,r.quo.low,r.rem); */
  return r;
}


void print_rec(FILE *fp,counter x)
{
  div r;

  if (!zero(x)) {
    r = div10(x);
    print_rec(fp,r.quo);
    putc('0'+r.rem,fp);
  }
}

void print10(FILE *fp,counter x)
{
  if (zero(x))
    putc('0',fp);
  else
    print_rec(fp,x);
}

int read10(FILE *fp,counter *r)
{
  int c;

  *r = promote(0);
  while (!isdigit(c = getc(fp)))
         if (c == EOF)
           return c;
  do {
    *r = multi(*r,10);
    ADD(*r,c-'0');
  } while (isdigit(c = getc(fp)));

  return c;
}

counter addi(counter x,unsigned long i)
{
  ADD(x,i);
  return x;
}
