/********************************************************************/
/*                              Poly                                */
/*                                                                  */
/*   Luc Maranget, projet Moscova, Inria Rocquencourt               */
/*                                                                  */
/* Copyright 2005 Institut National de Recherche en Informatique et */
/* en Automatique.  All rights reserved.  This file is distributed  */
/* under the terms of the Q Public License version 1.0.             */ 
/********************************************************************/

#define NMAX 26
int n=10;

#include <stdio.h>
#include <stdlib.h>
#include<unistd.h>
#include<string.h>
#include<time.h>
#include "count.h"

counter count;

#ifndef FILENAME_MAX
#define FILENAME_MAX 128
#endif

typedef enum statut {libre, plein} statut;

typedef struct point {
  statut stat;
  struct point *next;
  struct point *voisins[5];
} point;

/* damier */


point damier[NMAX][2*(NMAX-1)];

int verbose = 0;

int check(int i,int j)
{
  if ((0 <= i) && (i < n) && 0 <= j && (j < 2*(n-1))) {
    return ((i > 0) || (j > n-2));
  } else {
   return 0;
  }
}

void set_nexts(int i,int j) {
  int k=0;
  point *r = &damier[i][j];

  if (check(i+1,j)) {r->voisins[k++] = &damier[i+1][j];}
  if (check(i,j+1)) {r->voisins[k++] = &damier[i][j+1];}
  if (check(i-1,j)) {r->voisins[k++] = &damier[i-1][j];}
  if (check(i,j-1)) {r->voisins[k++] = &damier[i][j-1];}
  r->voisins[k] = 0;
}

void init_damier(void)
{
  int i,j;

  for (i=0;i<n;i++) {
    for (j=0;j<2*(n-1);j++) {
      set_nexts(i,j);
    }
  }
}


void print_sol()
{
  int i,j;
  point *r;

  for (i=n-1;i >= 0;i--) {
    for (j=0;j<2*(n-1);j++) {
      r = &damier[i][j];
      putchar(r->stat==libre?'.':(r->stat==plein?'#':'X'));
    }
    putchar('\n');
  }
  putchar('\n');
}

/* enumerateur rapide */
void search(int k,point *nexts,int l) {
  point *news,**v,*me;

  if (k > 1) {
    
    for (;nexts;) {
      v = &(nexts->voisins[0]);
      nexts = nexts->next; l--;
      news = nexts;
      for (me = *v ; me ; me = *++v) {
        if ((me)->stat == libre) {
          (me)->stat = plein;
          (me)->next = news;
          news = me; l++;
        }
      }
      search(k-1,news,l);
      for (; news != nexts; news = news->next) {
        news->stat = libre; l--;
      }
    }
  } else {
    l = l * (l-1)/2;
    for (;nexts;) {
      v = &(nexts->voisins[0]);
      nexts = nexts->next;
      for (me = *v ; me ; me = *++v) {
        if (me->stat == libre)
          l++;
      }
    }
    ADD(count,l);
  }
}

/* donnees pour repartir */
int seen[NMAX];
#ifndef SAVE_NOW
#define SAVE_NOW 15
#endif

char outfile[FILENAME_MAX];

#define MAGIC 1234

unsigned long magic=MAGIC;

void save_now()
{
  char tmp[FILENAME_MAX];
  FILE *fp;
  size_t wrote=0;
  int save_level=SAVE_NOW;

  if (! *outfile)
    return;
  sprintf(tmp,"%s.%d",outfile,getpid());
  fp = fopen(tmp,"wb");
  if (!fp) {
    perror("Cannot open tmp file");
    fprintf(stderr,"tmp file name is %s\n",tmp);
    return;
  }
  wrote += fwrite(&magic,sizeof(magic),1,fp);
  wrote += fwrite(&save_level,sizeof(save_level),1,fp);
  wrote += fwrite(&n,sizeof(n),1,fp);
  wrote += fwrite(&count,sizeof(count),1,fp);
  wrote += fwrite(seen,sizeof(seen[0]),n,fp);
  if (wrote != 3 + 1 + n) {
    perror("Cannot write to tmp file");
    fprintf(stderr,"tmp file name is %s\n",tmp);
    fclose(fp);
    return;
  }
  if (fclose(fp)) {
    perror("Cannot write to tmp file");
    return;
  }
  if (rename(tmp,outfile)) {
    perror("Cannot rename tmp file");
    return;
  }
  if (verbose > 0) {
    time_t buff;

    time(&buff);
    fprintf(stderr,"State saved in %s : %s",outfile,ctime(&buff));
  }
}

void restore_now(char *check)
{
  FILE *fp;
  int read;
  int save_level;

  fp = fopen(check,"rb");
  if (!fp) {
    fprintf(stderr,"Cannot open file %s: ",check);
    perror("");
    exit(-1);
  }
  read = 0;
  while (read < 1) {
    read += fread(&magic,sizeof(magic),1,fp);
    if (feof(fp) || ferror(fp)) {
      perror("Error reading checkpoint file");
      exit(-1);
    }
  }
  if (magic != MAGIC) {
    fprintf(stderr,"Bad magic number\n");
    exit(-1);
  }
  read = 0;
  while (read < 1) {
    read += fread(&save_level,sizeof(save_level),1,fp);
    if (feof(fp) || ferror(fp)) {
      perror("Error reading checkpoint file");
      exit(-1);
    }
  }
  if (save_level != SAVE_NOW) {
    fprintf(stderr,"Save levels are different in checkpoint file (%d) and in program (%d)\n",save_level,SAVE_NOW);
    exit(-1);
  }

  read=0;
  while (read < 1) {
    read += fread(&n,sizeof(n),1,fp);
    if (feof(fp) || ferror(fp)) {
      perror("Error reading checkpoint file");
      exit(-1);
    }
  }
  if (n < SAVE_NOW || n > NMAX) {
    fprintf(stderr,"Depth %d is not correct (%d <= n < %d)\n",n,SAVE_NOW,NMAX);
    exit(-1);
  }

  read = 0;
  while (read < 1) {
    read += fread(&count,sizeof(count),1,fp);
    if (feof(fp) || ferror(fp)) {
      perror("Error reading checkpoint file");
      exit(-1);
    }
  }

  read = 0;
  while (read < n) {
    read += fread(seen,sizeof(seen[0]),n,fp);
    if (feof(fp) || ferror(fp)) {
      perror("Error reading checkpoint file");
      exit(-1);
    }
  }
  if (getc(fp) != EOF) {
    fprintf(stderr,"Error reading checkpoint file: file too long\n");
    exit(-1);
  }
  fclose(fp);

  if (verbose) {
    fprintf(stderr,"Restored checkpoint file for n=%d, count=",n);
    print10(stderr,count);
    fprintf(stderr,"\n");
  }
}

void restart(int k, point *nexts,int size)
{
  point *me,*news,**p;
  unsigned int s;


  if (k == SAVE_NOW) {
    save_now() ;
    search(k,nexts,size);
    return;
  }
  if (seen[k] > 0) {
    for (s = seen[k] ; s > 1; s--) {/* deja fait */
      nexts = nexts->next;
      size--;
    }
    me = nexts;
    nexts = nexts->next;
    size--;
    me->stat = plein;
    news = nexts;
    for (p=&(me->voisins[0]) ; *p ; p++) {
      if ((*p)->stat == libre) {
        (*p)->stat = plein;
        (*p)->next = news; size++;
        news = *p;
      }
    }
    restart(k-1,news,size);
    for (; news != nexts; news = news->next) {
      news->stat = libre;
      size--;
    }
  }
  for (;nexts;) {
    seen[k]++;
    me = nexts;
    nexts = nexts->next;
    size--;
    news = nexts;
    for (p=&(me->voisins[0]) ; *p ; p++) {
      if ((*p)->stat == libre) {
        (*p)->stat = plein;
        (*p)->next = news; size++;
        news = *p;
      }
    }
    restart(k-1,news,size);
    for (; news != nexts; news = news->next) {
      news->stat = libre;
      size--;
    }
  }
  seen[k] = 0;
}

void usage(char *pgm)
{
  fprintf(stderr,"usage: %s [-r checkpointfile] [-o checkpointfile] [-v]* [n]\n",pgm);
  exit(-1);
}


int main(int argc,char **argv)
{
  point *p;
  char *arg=0;
  char *pgm = argv[0];
  int restored = 0;
  int do_save = 0;

  count = promote(0);

  while (--argc) {
    arg = *++argv;
    if (*arg == '-') {
      if (*(++arg) == 'r') {
        --argc;
        arg = *++argv;
        if (arg) {
          restore_now(arg);
          restored = 1;
        } else
          usage(pgm);
      } else if (*arg == 'o') {
        do_save = 1;
        --argc;
        arg = *++argv;
        if (arg) {
          strcpy(outfile,arg);
        } else
          usage(pgm);
      } else
        if (*arg == 'v') {
          verbose++;
        } else
          usage(pgm);
    } else {
      break;
    }
  }
  if (argc == 1) {
    if (restored && verbose)
      fprintf(stderr,"%s: warning, numerical argument ignored\n",pgm);
    else
      n = atoi(arg);
  } else if (argc != 0) {
    usage(pgm);
  }
  if (n > NMAX) {
    fprintf(stderr,"argument must be <= %d\n",NMAX);
    exit(2);
  }

  init_damier();

  p=&damier[0][n-2];
  p->next = 0; p->stat = plein;

  if (n > SAVE_NOW && (restored || *outfile))
    restart(n-1,p,1);
  else {
    if (do_save && *outfile) {
      fprintf(stderr,\
            "%s: warning, I will save no checkpoint files,(n=%d <= %d)\n",\
            pgm,n,SAVE_NOW);
    }
    search(n-1,p,1);
  }

  printf("%d\t",n);
  print10(stdout,count);
  putchar('\n');

  return 0;
}
